﻿using System;
using System.Runtime.InteropServices;
using System.Windows;
using System.Windows.Interop;

namespace TestProject
{
    /// <summary>
    /// 세션 통지 헬퍼
    /// </summary>
    public class SessionNotificationHelper : IDisposable
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Import
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Private

        #region WTS 세션 통지 등록하기 - WTSRegisterSessionNotification(windowHandle, flag)

        /// <summary>
        /// WTS 세션 통지 등록하기
        /// </summary>
        /// <param name="windowHandle">윈도우 핸들</param>
        /// <param name="flag">플래그</param>
        /// <returns>처리 결과</returns>
        [DllImport("wtsapi32")]
        private static extern bool WTSRegisterSessionNotification(IntPtr windowHandle, int flag);

        #endregion
        #region WTS 세션 통지 등록 취소하기 - WTSUnRegisterSessionNotification(windowHandle)

        /// <summary>
        /// WTS 세션 통지 등록 취소하기
        /// </summary>
        /// <param name="windowHandle">윈도우 핸들</param>
        /// <returns>처리 결과</returns>
        [DllImport("wtsapi32")]
        private static extern bool WTSUnRegisterSessionNotification(IntPtr windowHandle);

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Event
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 세션 변경시 이벤트 - SessionChanged

        /// <summary>
        /// 세션 변경시 이벤트
        /// </summary>
        public event EventHandler<SessionNotificationEventArgs> SessionChanged;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// NOTIFY_FOR_THIS_SESSION
        /// </summary>
        private const int NOTIFY_FOR_THIS_SESSION = 0;

        /// <summary>
        /// SESSION_CHANGED
        /// </summary>
        private const int SESSION_CHANGED = 0x02B1;

        /// <summary>
        /// SESSION_LOCK
        /// </summary>
        private const int SESSION_LOCK = 0x7;

        /// <summary>
        /// SESSION_UNLOCK
        /// </summary>
        private const int SESSION_UNLOCK = 0x8;

        /// <summary>
        /// 등록 여부
        /// </summary>
        private bool registered = false;

        /// <summary>
        /// 윈도우 인터롭 헬퍼
        /// </summary>
        private WindowInteropHelper windowInteropHelper;

        /// <summary>
        /// 세션 잠금 마지막 시간
        /// </summary>
        private DateTime lastTimeSessionLocked = DateTime.MinValue;

        /// <summary>
        /// 세션 잠금 취소 마지막 시간
        /// </summary>
        private DateTime lastTimeSessionUnlocked = DateTime.MinValue;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region 이벤트 발생 가능 여부 - EnableRaisingEvents

        /// <summary>
        /// 이벤트 발생 가능 여부
        /// </summary>
        protected bool EnableRaisingEvents
        {
            get
            {
                return this.registered;
            }
            set
            {
                bool haveXp =  Environment.OSVersion.Platform == PlatformID.Win32NT &&
                              (Environment.OSVersion.Version.Major > 5 ||
                              (Environment.OSVersion.Version.Major == 5 &&
                               Environment.OSVersion.Version.Minor >= 1));

                if(!haveXp)
                {
                    this.registered = false;

                    return;
                }

                if(value == true && !registered)
                {
                    WTSRegisterSessionNotification(this.windowInteropHelper.Handle, NOTIFY_FOR_THIS_SESSION);
 
                    this.registered = true;
                }
                else if(value == false && registered)
                {
                    WTSUnRegisterSessionNotification(this.windowInteropHelper.Handle);

                    this.registered = false;
                }
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - SessionNotificationHelper(window)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="window">윈도우</param>
        public SessionNotificationHelper(Window window)
        {
            this.windowInteropHelper = new WindowInteropHelper(window);

            window.Loaded += window_Loaded;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public
        //////////////////////////////////////////////////////////////////////////////// Function

        #region 리소스 해제하기 - Dispose()

        /// <summary>
        /// 리소스 해제하기
        /// </summary>
        public void Dispose()
        {
            if(this.registered)
            {
                EnableRaisingEvents = false;
            }
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private
        //////////////////////////////////////////////////////////////////////////////// Event

        #region 윈도우 로드시 처리하기 - window_Loaded(sender, e)

        /// <summary>
        /// 윈도우 로드시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void window_Loaded(object sender, RoutedEventArgs e)
        {
            HwndSource windowHandleSource = HwndSource.FromHwnd(this.windowInteropHelper.Handle);

            windowHandleSource.AddHook(new HwndSourceHook(WndProc));

            EnableRaisingEvents = true;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Function

        #region 세션 잠금 처리하기 - ProcessSessionLock()

        /// <summary>
        /// 세션 잠금 처리하기
        /// </summary>
        private void ProcessSessionLock()
        {
            DateTime currentTime = DateTime.Now;

            if(this.lastTimeSessionLocked == DateTime.MinValue)
            {
                this.lastTimeSessionLocked = currentTime;
            }
            else
            {
                TimeSpan elapsed = currentTime - this.lastTimeSessionLocked;

                this.lastTimeSessionLocked = currentTime;

                if(elapsed.TotalMilliseconds < 100)
                {
                    return;
                }
            }

            if(SessionChanged != null)
            {
                SessionChanged(this, new SessionNotificationEventArgs(SessionNotification.Lock));
            }
        }

        #endregion
        #region 세션 잠금 취소 처리하기 - ProcessSessionUnLock()

        /// <summary>
        /// 세션 잠금 취소 처리하기
        /// </summary>
        private void ProcessSessionUnLock()
        {
            DateTime currentTime = DateTime.Now;

            if(this.lastTimeSessionUnlocked == DateTime.MinValue)
            {
                this.lastTimeSessionUnlocked = currentTime;
            }
            else
            {
                TimeSpan elapsed = currentTime - this.lastTimeSessionUnlocked;

                this.lastTimeSessionUnlocked = currentTime;

                if(elapsed.TotalMilliseconds < 100)
                {
                    return;
                }
            }

            if(SessionChanged != null)
            {
                SessionChanged(this, new SessionNotificationEventArgs(SessionNotification.Unlock));
            }
        }

        #endregion
        #region 윈도우 프로시저 처리하기 - WndProc(windowHandle, message, wordParameter, longParameter, handled)

        /// <summary>
        /// 윈도우 프로시저 처리하기
        /// </summary>
        /// <param name="windowHandle">윈도우 핸들</param>
        /// <param name="message">메시지</param>
        /// <param name="wordParameter">WORD 매개 변수</param>
        /// <param name="longParameter">LONG 매개 변수</param>
        /// <param name="handled">처리 여부</param>
        /// <returns>처리 결과</returns>
        private IntPtr WndProc(IntPtr windowHandle, int message, IntPtr wordParameter, IntPtr longParameter, ref bool handled)
        {
            if(message == SESSION_CHANGED)
            {
                if(wordParameter.ToInt32() == SESSION_LOCK)
                {
                    ProcessSessionLock();
                }
                else if(wordParameter.ToInt32() == SESSION_UNLOCK)
                {
                    ProcessSessionUnLock();
                }
            }
        
            return IntPtr.Zero;
        }

        #endregion
    }
}